#include "browsertreedbusextension.h"
#include "browser.h"
#include "browserutils.h"
#include <macros.h>
#include <unistd.h> // gethostname
#include <QtDebug>
#include <TAppInfoList>
#include <TApplicationsPoolProxy>
#include <TApplicationInterface>
#include <cumbiadbusplugin.h>
#include <QDir>
#include <QPluginLoader>

BrowserTreeDBusExtension::BrowserTreeDBusExtension(BrowserTree *tree) : QObject(tree)
{
    m_tree = tree;
    m_qudbusctrl_i = NULL;
    m_loadCumbiaPlugin();
    if(m_qudbusctrl_i)
        m_qudbusctrl_i->addCtrlListener(this);
    TApplicationsPoolProxy *appp = new TApplicationsPoolProxy(this);
    connect(appp, SIGNAL(serviceRegistered(const TAppInfo&)), this, SLOT(newApplicationRegisteredToBus(const TAppInfo&)));
    connect(appp, SIGNAL(serviceUnregistered(const TAppInfo&)), this, SLOT(applicationUnregisteredFrom(const TAppInfo&)));
}

BrowserTreeDBusExtension::~BrowserTreeDBusExtension()
{
    if(m_qudbusctrl_i)
        delete m_qudbusctrl_i;
    delete findChild<TApplicationsPoolProxy *>();
}

void BrowserTreeDBusExtension::setApplicationsRunning(const QList<AppInfo> &ail)
{
    qDebug() << __FUNCTION__ << "list " << ail.size();
    for(int i = 0; i < ail.size(); i++)
        m_newApplicationLaunched(ail.at(i));
}

bool BrowserTreeDBusExtension::isCumbiaApp(const QString &exename) const
{
    if(!m_qudbusctrl_i) return false;
    qDebug() << __FUNCTION__ << "searching dbus services containing" << exename;
    QList<QuAppInfo> ail = m_qudbusctrl_i->findApps(QStringList()); // find all apps
    foreach(const QuAppInfo& ai, ail) {
        qDebug() << __FUNCTION__ << "service " << ai.dbusServiceName();
        if(ai.dbusServiceName().contains("quapplication") && ai.dbusServiceName().contains(exename, Qt::CaseInsensitive))
            return true;
    }
    return false;
}

void BrowserTreeDBusExtension::m_newApplicationLaunched(const AppInfo &ai)
{
    QString anam, dbus_service_name;
    QStringList args;
    m_extractInfo(ai, anam, args, dbus_service_name);

    printf("\e[1;32mBrowserTreeDBusExtension::m_newApplicationLaunched: %s [%s] dbusServiceName %s\e[0m\n",
           qstoc(anam), qstoc(args.join(" ")), qstoc(dbus_service_name));

    /* get EApplicationLauncher associated to executable and arguments */
    EApplicationLauncher *launcher = m_tree->GetAppByExeAndArgs(anam, args);
    if(launcher)
    {
        printf("\e[1;32m m_newApplicationLauncehr Founnd NEW launcher! %s\e[0m\n", qstoc(launcher->CompleteCommand()));
        launcher->AppUtils()->SetAppliStatus(RUNNING);
        launcher->XUtils()->SetX11Status(EXISTING);
        launcher->setProperty("dbusExtensionManaged", true);
        launcher->setProperty("is_cumbia", ai.isQAppI());
        m_tree->ApplicationRaised(true, launcher);
        m_tree->expand(launcher);
        m_tree->setDBusService(launcher, dbus_service_name);
        if(ai.isQAppI() && m_qudbusctrl_i)
            m_qudbusctrl_i->start_monitor(ai.qai.dbusServiceName());
    }
}

void BrowserTreeDBusExtension::m_applicationClosed(const AppInfo& ai)
{
    QString anam, dbus_service_name;
    QStringList args;
    m_extractInfo(ai, anam, args, dbus_service_name);
    EApplicationLauncher *launcher = m_tree->GetAppByExeAndArgs(anam, args);
    if(launcher)
    {
        qprintf("\e[1;32m m_applicationClosed: Found launcher! %s (for closing)\e[0m\n", qstoc(launcher->CompleteCommand()));
        launcher->AppUtils()->SetAppliStatus(NOT_RUNNING);
        launcher->XUtils()->SetX11Status(NOT_EXISTING);
        launcher->setProperty("dbusExtensionManaged", true);
        launcher->setProperty("is_cumbia", ai.isQAppI());
        m_tree->ApplicationExited(false, launcher);
    }
}

void BrowserTreeDBusExtension::onAppRegistered(const QuAppInfo &qai)
{
    printf("\e[0;36mBrowserTreeDBusExtension::onAppRegistered (cumbia) %s %s [%s] display %s num %d screen %d\e[0m\n", qstoc(qai.exename()),
           qstoc(qai.args().join(" ")), qstoc(qai.dbusServiceName()), qstoc(qai.display_host()), qai.display_number(),
           qai.screen_number());
}

void BrowserTreeDBusExtension::onAppUnregistered(const QuAppInfo &qai)
{
    printf("\e[1;36mBrowserTreeDBusExtension::onAppUnregistered (cumbia) %s %s [%s] display %s num %d screen %d\e[0m\n", qstoc(qai.exename()),
           qstoc(qai.args().join(" ")), qstoc(qai.dbusServiceName()), qstoc(qai.display_host()), qai.display_number(),
           qai.screen_number());
    AppInfo ai(qai);
    if(m_quapp_in_this_display(qai)) {
        m_applicationClosed(ai);
        if(m_qudbusctrl_i)
            m_qudbusctrl_i->stop_monitor(qai.dbusServiceName());
    }
}

AppInfo BrowserTreeDBusExtension::findApp(const QString& commandLine)
{
    AppInfo ret;
    TApplicationsPoolProxy* appp = findChild<TApplicationsPoolProxy *>();
    QStringList full_args = commandLine.split(QRegExp("\\s+"), QString::SkipEmptyParts);
    QStringList args(full_args);
    QString exe;
    if(args.size())
    {
        exe = args.first();
        if(exe.contains("/"))
            exe = exe.split("/").last();
        args.removeFirst(); /* in args leave the command line parameters */
    }

    QList<TAppInfo> ail;
#if QTANGO_VERSION < 0x060406
    // list of applications with name in ailwn...
    QList<TAppInfo> ailwn = appp->findApplicationsWithName(exe);
    foreach(const TAppInfo& ai, ailwn) {
        // ... name && args in ail
        if(ai.args == args)
            ail << ai;
    }
#else
    // findApplications: introduced in QTango 6.4.6
    ail = appp->findApplicationsByNameArgs(exe, args);
#endif
    foreach(const TAppInfo& ai, ail) {
        if(m_qtango_app_in_this_display(ai)) {
            ret.tai = ai;
            break;
        }
    }
    // cumbia apps
    if(m_qudbusctrl_i)
    {
        QList<QuAppInfo> ail = m_qudbusctrl_i->findApps(full_args);
        foreach(const QuAppInfo &ai, ail) {
            if(m_quapp_in_this_display(ai)) {
                ret.qai = ai;
                break;
            }
        }
    }
    return ret;
}

void BrowserTreeDBusExtension::m_loadCumbiaPlugin()
{
    QDir pluginsDir(QUMBIA_PLUGIN_DIR);
    // pluginsDir.cd("plugins");
    foreach (QString fileName, pluginsDir.entryList(QDir::Files)) {
        QPluginLoader pluginLoader(pluginsDir.absoluteFilePath(fileName));
        QObject *plugin = pluginLoader.instance();
        if (plugin) {
            QuDBusPluginInterface *dpi = qobject_cast<QuDBusPluginInterface *>(plugin);
            if(dpi)
                m_qudbusctrl_i = dpi->getAppCtrlIface();
        }
        else
            perr("BrowserTreeDBusExtension.m_loadCumbiaPlugin: error loading plugin: %s", qstoc(pluginLoader.errorString()));
    }
    if(m_qudbusctrl_i)
        printf("\e[1;32m* \e[0msuccessfully loaded the \e[1;32;4mcumbia dbus\e[0m plugin\n");
}

void BrowserTreeDBusExtension::m_extractInfo(const AppInfo &ai, QString &anam, QStringList &args, QString &dbus_service_name) const
{
    if(ai.isQAppI()) {
        anam = ai.qai.exename();
        args = ai.qai.args();
        dbus_service_name = ai.qai.dbusServiceName();
    }
    else {
        anam = ai.tai.appName;
        args = ai.tai.args;
        dbus_service_name = ai.tai.dbusServiceName;
    }
}

bool BrowserTreeDBusExtension::dbusRaiseCumbia(const QString& commandLine)
{
    AppInfo ai;
    bool this_display = false;
    if(m_qudbusctrl_i)
    {
        ai = findApp(commandLine);
        if(ai.isQAppI()) {
            this_display = m_quapp_in_this_display(ai.qai);
            if(this_display)
                m_qudbusctrl_i->raise(ai.qai);
        }
    }
    return !ai.isEmpty() && ai.isQAppI() && this_display;
}

void BrowserTreeDBusExtension::dbusStopQTangoCumbia(const QString& commandLine)
{
    qDebug() << "X dbus stop!" << commandLine;
    AppInfo ai = findApp(commandLine);
    QStringList full_args = commandLine.split(QRegExp("\\s+"), QString::SkipEmptyParts);
    if(ai.isTAppI() && m_qtango_app_in_this_display(ai.tai)) {
        TApplicationInterface aif(ai.tai.dbusServiceName, ai.tai.dbusServicePath, QDBusConnection::sessionBus());
        aif.quit();
    }
    else if(m_qudbusctrl_i && full_args.size() > 0) // cumbia apps
    {
        full_args[0] = full_args[0].section('/', -1);
        QList<QuAppInfo> ail = m_qudbusctrl_i->findApps(full_args);
        foreach(QuAppInfo ai, ail) {
            if(m_quapp_in_this_display(ai))
                m_qudbusctrl_i->close(ai);
        }
    }
}

void BrowserTreeDBusExtension::newApplicationRegisteredToBus(const TAppInfo& appInfo)
{
    AppInfo ai(appInfo);
    if(m_qtango_app_in_this_display(appInfo))
        m_newApplicationLaunched(ai);
}

void BrowserTreeDBusExtension::applicationUnregisteredFrom(const TAppInfo &appInfo)
{
    AppInfo ai(appInfo);
    printf("\e[1;35mBrowserTreeDBusExtension::applicationUnregisteredFrom: %s [%s] dbusServiceName %s "
           "dbusServicePath %s\e[0m\n", qstoc(appInfo.appName), qstoc(appInfo.args.join(" ")), qstoc(appInfo.dbusServiceName),
           qstoc(appInfo.dbusServicePath));
    if(m_qtango_app_in_this_display(appInfo))
        m_applicationClosed(ai);
}

/*!
 * \brief BrowserTreeDBusExtension::refreshList
 *
 * Queries the DBus to find QTango or cumbia registered applications.
 * The list is then filtered to discard those running on another host/display
 *
 * \return list of applications registered with DBus and running in this host/display
 *
 */
QList<AppInfo> BrowserTreeDBusExtension::refreshList()
{
    QList<AppInfo> ret;
    TApplicationsPoolProxy *tappPP = findChild<TApplicationsPoolProxy *>();
    TAppInfoList tail = tappPP->applicationsList();
    for(int i = 0; i < tail.size(); i++) {
        printf("\e[1;33mBrowserTreeDBusExtension::refreshList: checking if %s in this display\e[0m\n", qstoc(tail.at(i).dbusServiceName));
        if(m_qtango_app_in_this_display(tail.at(i))) {
            printf("YES IT IS\n");
            ret << AppInfo(tail.at(i));
        }
    }
    if(m_qudbusctrl_i) // cumbia apps
    {
        QList<QuAppInfo> ail = m_qudbusctrl_i->findApps(QStringList()); // finds all apps
        foreach (QuAppInfo ai, ail) {
            if(m_quapp_in_this_display(ai)) {
                printf("BrowserTreeDBusExtension::refreshList\e[0;32;4m ### app %s %s is running on this display\e[0m\n", qstoc(ai.exename()),
                       qstoc(ai.args().join(" ")));
                ret << AppInfo(ai);
            }
            else {
                printf("BrowserTreeDBusExtension::refreshList\e[0;35;4m ### app %s %s is NOT running on this display\e[0m\n", qstoc(ai.exename()),
                       qstoc(ai.args().join(" ")));
            }
        }
    }
    return ret;
}

bool BrowserTreeDBusExtension::m_quapp_in_this_display(const QuAppInfo &ai)
{
    BrowserUtils bu;
    printf("\e[0;33mBrowserTreeDBusExtension::m_quapp_in_this_display app %s %s\n"
           "*\tthis host %s dat host %s\n"
           "*\tthis disp %d dat disp %d\n"
           "*\t this screen %d dat screen %d\e[0m\n", qstoc(ai.exename()), qstoc(ai.args().join(" ")),
           qstoc(bu.display_host()), qstoc(ai.display_host()), bu.display_number(), ai.display_number(),
           bu.screen_number(), ai.screen_number());
    return ai.display_host() == bu.display_host() && ai.display_number() == bu.display_number()
            && ai.screen_number() == bu.screen_number();
}

bool BrowserTreeDBusExtension::m_qtango_app_in_this_display(const TAppInfo &ai)
{
    const QString& dbus_sn = ai.dbusServiceName;
    // from qtangocore/src/tapplication/tapplication_p
    // char hostname[HOST_NAME_MAX];
    // QTango uses gethostname()
    // if(gethostname(hostname, HOST_NAME_MAX))
    // QString qsdisplay;
    // qsdisplay = QString(display);
    // qsdisplay.remove(QRegExp("\\.\\d*")).remove(":");
    // m_dbusServiceName = "it.trieste.elettra." + QString(hostname) + "_" + qsdisplay  +
    //         ".TApplications." + composeDbusObjectName() + ".pid_" + QString::number(getpid());
    //
    char hostname[256];
    gethostname(hostname, 256);
    BrowserUtils bu;
    QString this_host = QString(hostname);
    QString qsdisplay = QString(getenv("DISPLAY"));
    qsdisplay.remove(QRegExp("\\.\\d*")).remove(":");
    QString host_disp = QString(hostname) + "_" + qsdisplay; // as composed in QTango tapplication_p
    // it\.trieste\.elettra\.(.*)\.TApplications\..*
    // capture everything between it.trieste.elettra and TApplications
    QRegExp qtango_dbus_service_re("it\\.trieste\\.elettra\\.(.*)\\.TApplications\\..*");
    if(qtango_dbus_service_re.indexIn(dbus_sn) > -1) {
        QStringList ct = qtango_dbus_service_re.capturedTexts();
        printf("\e[0;35mBrowserTreeDBusExtension::m_qtango_app_in_this_display app \"%s\" \"%s\"\n"
               "*\tthis host/display info %s dat host/display info %s\n"
               "*\tthis disp %d dat disp %d\e[0m\n", qstoc(ai.appName), qstoc(ai.args.join(" ")),
               qstoc(host_disp), qstoc(ct.at(1)));
        return ct.size() > 1 && ct.at(1) == host_disp;
    }
    else {
        printf("\e[0;31mBrowserTreeDBusExtension::m_qtango_app_in_this_display app \"%s\" \"%s\"\n"
               "*\t service \"%s\" does not match \"%s\"\e[0m\n",  qstoc(ai.appName), qstoc(ai.args.join(" ")),
               qstoc(dbus_sn), qstoc(qtango_dbus_service_re.pattern()));
    }
    return false;
}



